import argparse
import multiprocessing as mp
import sys
import time
from functools import partial
from pathlib import Path
from shutil import copytree

from .filesmanager import generate_filename, get_input_images, ignore_files
from .processing import process_image


def run(args=None):
    """
    Main CLI function.

    :param list[str] args: A list of arguments to be passed to the CLI.
    """
    if not args:
        args = sys.argv[1:]  # pragma: no cover

    parser = argparse.ArgumentParser(
        description="""ProperJPG aims to be a fast tool to convert and optimize any image before using it
        on the web. It does so by automatically stripping some meta data and also allowing easy resizing
        of a single or multiple images at once.
        
        Type 'properjpg -h' for help."""
    )
    parser.add_argument(
        "input", type=Path, help="The path to the file or folder to be used as input."
    )
    parser.add_argument(
        "output",
        type=Path,
        nargs="?",
        default=None,
        help="The path to the file or folder to save the output file(s).",
    )
    parser.add_argument(
        "-d",
        "--directory",
        action="store_true",
        help="""Acceps directories. Copy the original directory tree,
        search for all image files and tries to convert and save them to the selected output folder.
        In case no output folder is provided, "PROPER JPG" is created in the parent of the input folder.""",
    )
    parser.add_argument(
        "-wi",
        "--max-width",
        type=int,
        default=0,
        help="Sets the max width. Output images will be resized to this value or less.",
    )
    parser.add_argument(
        "-he",
        "--max-height",
        type=int,
        default=0,
        help="Sets the max height. Output images will be resized to this value or less.",
    )
    parser.add_argument(
        "-q",
        "--quality",
        type=int,
        default=0,
        help="""[ALPHA] If set, the input will be compressed to the set value (using Pillow library).
        
        ATTENTION: This is being tested. It may behave unpredictably.""",
    )
    args = parser.parse_args(args)
    input_path: Path = args.input

    max_width: int = args.max_width
    max_height: int = args.max_height
    quality: int = args.quality

    if args.output is None:
        if args.directory:
            output_path = input_path.parent.joinpath("PROPER JPG")
        else:
            output_path = generate_filename(input_path)
    else:
        output_path: Path = args.output

    print("Output path: ", output_path.resolve())

    if args.directory:
        print("STARTING PROCESS")
        start_time = time.perf_counter()

        try:
            assert input_path.is_dir()
        except:
            raise ValueError(
                f"The input path '{input_path.resolve()}' is not a directory."
            )

        try:
            assert not output_path.is_file()
        except:
            raise ValueError(
                f"The output path '{output_path.resolve()}' is not a directory."
            )

        copytree(input_path, output_path, ignore=ignore_files)

        image_list = [
            (image, save_path)
            for image, save_path in get_input_images(input_path, output_path)
        ]

        with mp.Pool() as pool:
            pool.starmap(
                partial(
                    process_image,
                    max_width=max_width,
                    max_height=max_height,
                    quality=quality,
                ),
                image_list,
            )

        end_time = time.perf_counter()
        print("")
        print("DONE")
        print(f"Elapsed time: {end_time - start_time}")
        print(f"Images processed: {len(image_list)}")

    else:
        print("STARTING PROCESS")
        start_time = time.perf_counter()

        try:
            assert input_path.is_file()
        except:
            raise ValueError(f"The input path '{input_path.resolve()}' is not a file.")

        process_image(input_path, output_path, max_width, max_height, quality)

        end_time = time.perf_counter()
        print("")
        print("DONE")
        print(f"Elapsed time: {end_time - start_time}")

    print("Output path: ", output_path.resolve())



import mimetypes
from functools import cache
from pathlib import Path

from PIL import Image


def process_image(
    image_path: Path,
    output_path: Path,
    max_width: int = 0,
    max_height: int = 0,
    quality: int = 0,
) -> None:
    """
    Process the image.

    :param image_path: The original image path.
    :param output_path: The path where the processed image should be saved.
    :param max_width: The max width the processed image can have.
    :param max_height: The max height the processed image can have.
    :param quality: The quality when optimizing an image.
    """
    print("Processing image: ", image_path)
    with Image.open(image_path) as image:

        if max_width != 0 or max_height != 0:
            ## If width or height is set.
            image = resize_image(image, max_width, max_height)

        if not "jpg" in mimetypes.guess_type(image_path):
            image = image.convert("RGB")
        if quality != 0:
            ## If Image quality is set.
            image.save(output_path.with_suffix(".jpg"), quality=quality, optimize=True)
        else:
            image.save(output_path.with_suffix(".jpg"))


def resize_image(
    image: Image.Image, max_width: int = 0, max_height: int = 0
) -> Image.Image:
    """
    Resize an image and return it.

    :param image: A Pill.Image object.
    :param max_width: The max width the processed image can have.
    :param max_height: The max height the processed image can have.
    """
    old_width, old_height = image.size[0], image.size[1]

    @cache
    def get_proper_sizes(
        max_width: int, max_height: int, old_width: int, old_height: int
    ) -> tuple[int, int]:

        new_width, new_height = max_width, max_height

        if max_width == 0:
            """If width is not set."""
            new_width = round(old_width * (max_height / old_height))
        elif max_height == 0:
            """If height is not set."""
            new_height = round(old_height * (max_width / old_width))
        else:
            if old_width > old_height:
                """If image's original width is bigger than original height."""
                new_height = round(old_height * (new_width / old_width))
            elif old_height > old_width:
                """If image's original height is bigger than original width."""
                new_width = round(old_width * (new_height / old_height))
            elif old_width == old_height:
                """If image's original width and height are the same."""
                if max_width > max_height:
                    """If new width is bigger than new height."""
                    new_width = max_height
                elif max_height > max_width:
                    """If new height is bigger than new width."""
                    new_height = max_height

        if new_width > max_width and max_width != 0:
            new_width = max_width
            new_height = round(old_height * (new_width / old_width))
        if new_height > max_height and max_height != 0:
            new_height = max_height
            new_width = round(old_width * (new_height / old_height))

        new_width = new_width
        new_height = new_height

        return new_width, new_height

    new_width, new_height = get_proper_sizes(
        max_width, max_height, old_width, old_height
    )

    image = image.resize((new_width, new_height), Image.LANCZOS)
    return image



import mimetypes
import os
from pathlib import Path


def ignore_files(dir: str, files: list[str]):
    """
    Returns a list of files to ignore.

    To be used by shutil.copytree()
    """
    return [f for f in files if Path(dir, f).is_file()]


def get_input_images(input_folder: Path, output_path: Path):
    """
    Get all images from a folder and it's subfolders.

    Also outputs a save path to be used by the image.

    :param input_folder: The folder to be scanned.
    :param output_path: The root folder of the destination path.
    """
    for root, _, files in os.walk(input_folder):
        for file in files:
            mime_type = mimetypes.guess_type(file)[0]

            if isinstance(mime_type, str):
                if "image" in mime_type:
                    image = Path(root, file)
                    relative_path = image.relative_to(input_folder)
                    save_path = Path(output_path, relative_path)
                    yield image, save_path


def generate_filename(input_path: Path) -> Path:
    gen_counter = 1
    gen_output = input_path.with_name(f"{input_path.stem}-{gen_counter}").with_suffix(
        ".jpg"
    )

    while gen_output.is_file():
        gen_counter += 1
        gen_output = input_path.with_name(
            f"{input_path.stem}-{gen_counter}"
        ).with_suffix(".jpg")

    output_path = gen_output
    return output_path



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['properjpg']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.1,<10.0.0']

entry_points = \
{'console_scripts': ['properjpg = properjpg.cli:run']}

setup_kwargs = {
    'name': 'properjpg',
    'version': '0.1.1',
    'description': 'Make any image ready for the web. Fast.',
    'long_description': 'ProperJPG\n=========\n\n*Make any image ready for the web. Fast.*\n\nProperJPG is a super fast, lightweight CLI app that converts images to jpg. It also resizes them!\n\n|\n\n**Actively soliciting contributors!**\n\nFeel free to open a pull request in case you find an issue or a way to improve the \napp. New features are also welcome, considering they don\'t add unecessary complexity to the\nuser experience.\n\nInstallation\n------------\n\n- `Regular`_\n- `Developer`_\n\nRegular\n~~~~~~~\n**Pipx (recommended)**\n   \n::\n\n   pipx install properjpg\n\n**Pip**\n\n::\n\n   pip install properjpg\n\nDeveloper\n~~~~~~~~~\n**Poetry**\n\n::\n\n   poetry add properjpg\n\n**Git**\n\n::\n\n   git clone https://github.com/vitorrloureiro/properjpg\n\nFeatures\n--------\n\n- `Smart resize`_\n- `Multiprocessing`_\n\nSmart resize\n~~~~~~~~~~~~\n\nIt has a super cool \'smart resize\' functionality.\nIt allows you to set a max width and height, and you can be sure that\nevery image compressed by it will be no larger or taller than what you\nspecify. This \'smart resize\' mode makes sure to don\'t resize images\nthat doesn\'t need to, and also takes in account if it\'s best for an\nimage to be resized based on it\'s width or height.\n\nMultiprocessing\n~~~~~~~~~~~~~~~\n\nThis app uses the multiprocessing module to leverage all the power on your computer.\nIt\'ll work faster if you have multiple cores.\n\nHow it works?\n-------------\n**This app works in two modes:**\n\n- `"Single file" mode`_\n- `"Directory" mode`_\n\n\n\n\n\n"Single file" mode\n~~~~~~~~~~~~~~~~~~\nInput an image path and the desired output path.\n\nUsage:\n   \n::\n\n   properjpg [input_path] [output_path] -wi=[max_width] -he=[max_height]\n\n\n"Directory" mode\n~~~~~~~~~~~~~~~~\nThis is where this app really shines. Input a directory path and a desired destination\npath and the app will clone the folder struct of the original directory on the output path.\nThen it will look for all images in the input folder and will try to convert (and\nresize, if you setted it to) them.\n\nProperJPG uses multiprocessing to speed up the process.\n\nUsage:\n\n::\n\n   properjpg [input_path] [output_path] -d -wi=[max_width] -he=[max_height]\n\n\nNotes\n-----\nThis software is in Alpha stage. A lot of things may change, including syntax and dependencies. I\'m looking for help\nto improve this tool in terms of speed, features and code readability. Feel free to make suggestions and appoint improvements!.\nAlso feel free to help me improve the tests 😅\n\nThe goal is to always keep the code with 100% test coverage.\n\nContributing\n------------\n\nRequirements\n~~~~~~~~~~~~\n\nThis repository automatically lints code with flake8 and black, and also runs mypy\nand pytest. Pull requests must pass in all those tests.\n\n- `black <https://github.com/psf/black>`_\n- `flake8 <https://github.com/PyCQA/flake8>`_\n- `mypy <https://github.com/python/mypy>`_\n- `pytest <https://github.com/pytest-dev/pytest>`_\n\nRoadmap\n~~~~~~~\n\n- Improve testing.\n- Improve Docs.\n- Improve UI (maybe switch to Click?).\n- Improve Performance.\n\nKnown Issues\n~~~~~~~~~~~~\n\nClient\n......\n- None\n\nDev\n...\n- Tests are incomplete.\n\nLicense\n-------\n**MIT**\n',
    'author': 'Vitor Loureiro',
    'author_email': 'miseravel.cruller-0o@icloud.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vitorrloureiro/properjpg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
